#!/usr/bin/perl
use Math::Trig;
$PI = 3.1415926;

#This Perl script was develop for relaxation stress tensor to zero values or its predefined values in the Abinit.
#Action: changes of rprim parameters (translation vectors) according to values of stress tensor.
#The are some restrictions about using. The name of input file must be "abinit_config.in". The rprim must be set in this style:
#rprim
#1x 1y 1z  !!Only one space character must be between numbers and at the end!! see the code which is using a "split" function at the perl language:
#2x 2y 3z                            if ($read == 1.0) {($r1[1], $r1[2], $r1[3]) = split ' ', $line; $read = 2.0; next}
#3x 3y 3z
#The stress tensor is marked as: sig[1] = sigma11, sig[2] = sigma22, sig[3] = sigma33, sig[4] = sigma23, sig[5] = sigma31, sig[6] = sigma12.



#The main setting

$no_wave_first_set = 1; # 1 - will check if parameter "IRDWFK -1" is blocked at the first iteration cycle, 0 - cancel checking (nothing will be done)

$tol_sig[1] = 0.100; $tol_sig[2] = 0.100; $tol_sig[3] = 0.100; #stress tolerance (0.1 means that all components of the stress tensor must be plus or minus about
$tol_sig[4] = 0.100; $tol_sig[5] = 0.100; $tol_sig[6] = 0.100; #given values)

@du=(0, 30, 30, 30, 50, 50, 50); #These constants are affecting the convergence speed. The first item of array can be zero (It is not used)!

@s=(0, 0, 0, 0, 0, 0, 0);	#The values of the stress tensor after relaxation. For the free stress state please set all values to zero. The first part of array
                          #is not used = always must be zero.

#Some parts of the main relaxation procedure which can be computed before the main cycle start.
foreach $i (1..6) {$du[$i]=$du[$i]*0.0001}
foreach $i (1..6) {$top_tol[$i] = $s[$i] + $tol_sig[$i]}
foreach $i (1..6) {$btn_tol[$i] = $s[$i] - $tol_sig[$i]}

foreach $i (1..6) {print $top_tol[$i]."\n"}
foreach $i (1..6) {print $btn_tol[$i]."\n"}

#konstatny a nemenne promenne------------------------------
$wfk_funct = 0.0;
$max_relaxation = 30.0;
get_config_in();
change_config_in();
  $num_main_cycle = 0.0;
  system("rm -f complete_results.txt");
  while ($num_main_cycle < $max_relaxation)
  {
    if ($wfk_funct == 1.0)					{start_read_wave_function()}
    if (($wfk_funct == 0.0) && ($no_wave_first_set == 1))	{check_no_read_wave_at_first_cycle()}
    if (($wfk_funct > 0.0) && ($wfk_set == 1))			{copy_wave_function()}
    system ("mpirun -np 8 abinip<abinit_config.files>&log.txt");
    get_config_in();                              #read input file
    get_config_out_file_as_array();               #read output file
    get_output_results();                         #reading results
    get_difference_of_stress_and_energy();        #
    write_results();                              #write res.
    
    #Convergence criterium, if all tolerances is accomplished the relaxation cycle will stop.
    last if (($sig[1] < $top_tol[1]) && ($sig[1] > $btn_tol[1]) && ($sig[2] < $top_tol[2]) && ($sig[2] > $btn_tol[2]) && ($sig[3] < $top_tol[3]) && ($sig[3] > $btn_tol[3]) && ($sig[4] < $top_tol[4])  && ($sig[4] > $btn_tol[4]) && ($sig[5] < $top_tol[5])  && ($sig[5] > $btn_tol[5]) && ($sig[6] < $top_tol[6])  && ($sig[6] > $btn_tol[6]));
    
    gues_new_vectors();			                      #run relaxation procedure
    change_config_in();                           #write new RPRIM to Abinit input file 
    system_change_for_next_step();                
    $num_main_cycle++;
    $wfk_funct++;
  }

#----------------procedures-------------------------------------------------------

#Open Abinit output file and copy all lines to array "@abinit_config_out". 
sub get_config_out_file_as_array {
  open(FILE_ABINIT_OUT,"<abinit_config.out");
  @abinit_config_out = <FILE_ABINIT_OUT>; close FILE_ABINIT_OUT;
}

#Getting the results from Abinit output file.
sub get_output_results {
  foreach $line (@abinit_config_out) {
    if ($line =~ /volume/)        {$volume =   substr($line, 26, 13)}
    if ($line =~ /Total\ energy/) {$energy =   substr($line, 22, 21)} #Not important and now doesn't work.
    if ($poc_s == 4.0)            {$sig[3] = substr($line, 14, 15); $sig[6] = substr($line, 43, 15); $poc_s++} #sig 21
    if ($poc_s == 3.0)            {$sig[2] = substr($line, 14, 15); $sig[5] = substr($line, 43, 15); $poc_s++} #sig 31
    if ($poc_s == 2.0)            {$sig[1] = substr($line, 14, 15); $sig[4] = substr($line, 43, 15); $poc_s++} #sig 32
    if ($line =~ /(GPa)/)         {$pressure = substr($line, 53, 27); $poc_s = 2.0}
  }
  #foreach $i (1..6) {$sig[$i]=$sig[$i]*(-1.0)}
}

#Writing results to the specific file "complete_results.txt".
sub write_results {
  open (FILE_COMPLETE_RESULTS, ">>complete_results.txt");
  print FILE_COMPLETE_RESULTS "------step number ".$num_main_cycle." --------\n";
  print FILE_COMPLETE_RESULTS "Relative primitive vectors: \n";
  print FILE_COMPLETE_RESULTS $r1[1]."\t".$r1[2]."\t".$r1[3]."\n";
  print FILE_COMPLETE_RESULTS $r2[1]."\t".$r2[2]."\t".$r2[3]."\n";
  print FILE_COMPLETE_RESULTS $r3[1]."\t".$r3[2]."\t".$r3[3]."\n";
  print FILE_COMPLETE_RESULTS "System energy: ".$energy."\n";
  print FILE_COMPLETE_RESULTS "Volume: ".$volume."\n";
  print FILE_COMPLETE_RESULTS "Stress tenzor: ".$pressure."\n";
  printf FILE_COMPLETE_RESULTS (" %11.9f %11.9f\n",$sig[1], $sig[6]);
  printf FILE_COMPLETE_RESULTS (" %11.9f %11.9f\n",$sig[2], $sig[5]);
  printf FILE_COMPLETE_RESULTS (" %11.9f %11.9f\n",$sig[3], $sig[4]);
  #printf FILE_COMPLETE_RESULTS ("y axe shift: %10.8f\n", $delta_r2_y);
  #printf FILE_COMPLETE_RESULTS ("z axe shift: %10.8f\n", $delta_r3_z);
  print FILE_COMPLETE_RESULTS "delta_energy: ".$delta_energy."\n";
  #print FILE_COMPLETE_RESULTS "delta_sigma_22: ".$delta_sigma_22."\n";
  #print FILE_COMPLETE_RESULTS "delta_sigma_33: ".$delta_sigma_33."\n";
  print FILE_COMPLETE_RESULTS "-----------------End this step------------------\n\n";
  close FILE_COMPLETE_RESULTS;
}

#Reading parameters RPRIM from Abinit input file.
sub get_config_in {
my $read;
open (FILE_CONFIG_IN, "<abinit_config.in"); @abinit_config_in = <FILE_CONFIG_IN>; close FILE_CONFIG_IN;
foreach $line (@abinit_config_in) {
  if ($line =~ /rprim/){$read = 1.0; next}
  if ($read == 1.0) {($r1[1], $r1[2], $r1[3]) = split ' ', $line; $read = 2.0; next}
  if ($read == 2.0) {($r2[1], $r2[2], $r2[3]) = split ' ', $line; $read = 3.0; next}
  if ($read == 3.0) {($r3[1], $r3[2], $r3[3]) = split ' ', $line; $read = "end"; next}
  }
}

#-----------------------------------------------------------------------------------------------
#The main relaxation procedure.
sub gues_new_vectors {
  foreach $i (1..6) {$ds[$i]=$s[$i]-$sig[$i]}
  @e=(0, $ds[1]*$du[1]+1, $ds[2]*$du[2]+1, $ds[3]*$du[3]+1, $ds[4]*$du[4], $ds[5]*$du[5], $ds[6]*$du[6]);
  print "e: @e\n"; print "ds: @ds\n"; print "s: @s\n";
  @r1=(0, $r1[1]*$e[1]+$r1[2]*$e[6]+$r1[3]*$e[5], $r1[1]*$e[6]+$r1[2]*$e[2]+$r1[3]*$e[4], $r1[1]*$e[5]+$r1[2]*$e[4]+$r1[3]*$e[3]);
  @r2=(0, $r2[1]*$e[1]+$r2[2]*$e[6]+$r2[3]*$e[5], $r2[1]*$e[6]+$r2[2]*$e[2]+$r2[3]*$e[4], $r2[1]*$e[5]+$r2[2]*$e[4]+$r2[3]*$e[3]);
  @r3=(0, $r3[1]*$e[1]+$r3[2]*$e[6]+$r3[3]*$e[5], $r3[1]*$e[6]+$r3[2]*$e[2]+$r3[3]*$e[4], $r3[1]*$e[5]+$r3[2]*$e[4]+$r3[3]*$e[3]);
}

#0, $r1[1]*$e[1]+$r1[2]*$e[6]+$r1[3]*$e[5], $r1[1]*$e[6]+$r1[2]*$e[2]+$r1[3]*$e[4], $r1[1]*$e[5]+$r1[2]*$e[4]+$r1[3]*$e[3]
#0, $r2[1]*$e[1]+$r2[2]*$e[6]+$r2[3]*$e[5], $r2[1]*$e[6]+$r2[2]*$e[2]+$r2[3]*$e[4], $r2[1]*$e[5]+$r2[2]*$e[4]+$r2[3]*$e[3]
#0, $r3[1]*$e[1]+$r3[2]*$e[6]+$r3[3]*$e[5], $r3[1]*$e[6]+$r3[2]*$e[2]+$r3[3]*$e[4], $r3[1]*$e[5]+$r3[2]*$e[4]+$r3[3]*$e[3]

#-----------------------------------------------------------------------------------------------------------------------------------

#Changing the input files according to values of RPRIM.
sub change_config_in {
  my $write;
  open (FILE_CONFIG_CHANGE, "<abinit_config.in"); @abinit_config_in = <FILE_CONFIG_CHANGE>; close FILE_CONFIG_CHANGE;
  open (FILE_CONFIG_CHANGE, ">abinit_config.in");
  foreach $line (@abinit_config_in) {
    if ($line =~ /rprim/) {printf FILE_CONFIG_CHANGE "rprim\n"; $write = 1.0; next}
    if    ($write == 1.0){printf FILE_CONFIG_CHANGE ("%11.9f %11.9f %11.9f\n",$r1[1], $r1[2], $r1[3]); $write = 2.0; next}
    elsif ($write == 2.0){printf FILE_CONFIG_CHANGE ("%11.9f %11.9f %11.9f\n",$r2[1], $r2[2], $r2[3]); $write = 3.0; next}
    elsif ($write == 3.0){printf FILE_CONFIG_CHANGE ("%11.9f %11.9f %11.9f\n",$r3[1], $r3[2], $r3[3]); $write = "end"; next}
    else  {print FILE_CONFIG_CHANGE $line}
  }
  close FILE_CONFIG_CHANGE;
}

#Rename wave function file (see xxxxxx.files).
#Copy Abinit out file as abinit_config.outX where X is number of relaxation step.
#Delete previous wave function (from previous iteration).
sub system_change_for_next_step {
  $new_name = "abinit_config.out".$num_main_cycle;
  system("mv abinit_config.out $new_name");
  system("rm -f abinit_configi_WFK");
  if ($wfk_set == 0)	{system("mv abinit_configo_WFK abinit_configi_WFK")}
} 

#Reading wave function from previous computation. Only for second and more steps!
sub start_read_wave_function{
  open(FILE_IN_IRDWFK,"<abinit_config.in"); @file_in_IRDWFK = <FILE_IN_IRDWFK>; close FILE_IN_IRDWFK;
  open(FILE_IN_IRDWFK,">abinit_config.in");
  foreach $line_wfk (@file_in_IRDWFK){
   if   ($line_wfk =~ /irdwfk/){print FILE_IN_IRDWFK "irdwfk 1\n"}
   else                        {print FILE_IN_IRDWFK $line_wfk}
  }
  close FILE_IN_IRDWFK;
}

#Not important. Computing only the differences of selected stress tensor.
sub get_difference_of_stress_and_energy {
  $delta_energy = abs ((abs $last_energy) - (abs $energy));
  $delta_sigma_22 = abs ((abs $last_sigma_22) - (abs $sigma_22));
  $delta_sigma_33 = abs ((abs $last_sigma_33) - (abs $sigma_33));
  $last_energy = $energy;
  $last_sigma_22 = $sigma_22;
  $last_sigma_33 = $sigma_33;  
}

#Checking if parameter "irdwfk -1" is blocked at the first iteration cycle.
sub check_no_read_wave_at_first_cycle {

  open (FILE_ABINIT_CONFIG_IN, "<abinit_config.in"); @lines_in = <FILE_ABINIT_CONFIG_IN>; close FILE_ABINIT_CONFIG_IN;
  open (FILE_ABINIT_CONFIG_OUT, ">abinit_config.in");
    foreach $line (@lines_in){
      if ($line =~ /irdwfk/){
        print FILE_ABINIT_CONFIG_OUT "#irdwfk 1\n";
      }
      else {print FILE_ABINIT_CONFIG_OUT $line}
    }
  close FILE_ABINIT_CONFIG_OUT;
}
